# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

from .button import Button
from dataclasses import dataclass
from pyrogram import (CallbackQuery, CallbackQueryHandler, Client,
                      InlineKeyboardButton, Filters)
from typing import Any, Callable, List


@dataclass(eq=False, init=False, repr=True)
class BaseHandler:
    def get_menus(self) -> List['BaseMenu']:
        raise NotImplementedError

    def process_keyboard(self, keyboard: List[List[Button]],
                         callback_query_id: str) -> List[
                             List[InlineKeyboardButton]]:
        return [[InlineKeyboardButton(
            button.name,
            callback_data=button.button_id)
                 for button in row] for row in keyboard]

    def setup(self, client: Client):
        for menu in self.get_menus():
            client.add_handler(CallbackQueryHandler(
                pass_handler(menu.on_callback, self),
                Filters.callback_data(menu.menu_id)))


def pass_handler(func: Callable[[Client, Any], None],
                 handler: BaseHandler) -> Callable[[Client, Any], None]:
    def on_callback(client: Client, context):
        if isinstance(context, CallbackQuery):
            func(handler, client, context, {})
        else:
            func(handler, client, context)

    return on_callback



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

from .base_database import BaseDatabase
from .errors import DeleteError, ExpireError, SetError
from datetime import timedelta
from typing import Optional, Union
import redis


class RedisDatabase(BaseDatabase):
    encoding = 'utf-8'
    expire: Optional[Union[int, timedelta]] = 86400
    server: redis.Redis

    def __init__(self, server: redis.Redis,
                 encoding='utf-8',
                 expire: Optional[Union[int, timedelta]] = 86400):
        self.encoding = encoding
        self.expire = expire
        self.server = server

    def get(self, callback_query_id: str) -> Optional[str]:
        content = self.server.get(callback_query_id)
        return content.decode(self.encoding) if content else None

    def set(self, callback_query_id: str, data: str):
        if not self.server.set(callback_query_id, data):
            raise SetError

        if self.expire and not self.server.expire(callback_query_id,
                                                  self.expire):
            raise ExpireError

    def delete(self, callback_query_id: str):
        try:
            self.server.delete(callback_query_id)
        except redis.RedisError:
            raise DeleteError



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

from .base_handler import BaseHandler
from .button import Button
from .database.base_database import BaseDatabase
from dataclasses import dataclass
from json import JSONDecodeError
from pyrogram import (Client, CallbackQueryHandler,
                      CallbackQuery, InlineKeyboardButton)
from pyrogram.client.filters.filters import create
from typing import Any, Callable, List

try:
    import orjson as json # noqa
except ImportError:
    import json


@dataclass(eq=False, init=False, repr=True)
class ParameterizedHandler(BaseHandler):
    database: BaseDatabase

    def __init__(self, database: BaseDatabase):
        self.database = database

    def filter(self, menu_id: str):
        def func(flt, callback: CallbackQuery):
            parameters = callback.data.split()

            if parameters[1] != menu_id:
                return False

            result = self.database.get(parameters[0])

            if result:
                try:
                    content = json.loads(result)

                    if parameters[1] in content:
                        callback.parameters = content[parameters[1]]
                        callback.parameters[
                            'callback_query_id'] = parameters[0]
                        callback.parameters[
                            'menu_id'] = parameters[1]

                        if len(parameters) > 2:
                            callback.parameters['element_id'] = parameters[2]

                            if not callback.parameters['same_menu']:
                                callback.parameters[
                                    parameters[1] + "_id"] = parameters[2]

                        return True

                    return False
                except JSONDecodeError:
                    return False

            return False

        return create(func, "ParameterizedCallbackData")

    def process_keyboard(self, keyboard: List[List[Button]],
                         callback_query_id: str) -> List[
                             List[InlineKeyboardButton]]:
        content = {}

        for row in keyboard:
            for button in row:
                content[button.button_id] = {
                    'callback_query_id': str(callback_query_id),
                    **button.parameters}

        self.database.set(callback_query_id, json.dumps(content))

        return [[InlineKeyboardButton(
            button.name, " ".join(map(str,
                [callback_query_id, button.button_id,
                 button.element_id])).strip().rstrip())
                 for button in row] for row in keyboard]

    def setup(self, client: Client):
        for menu in self.get_menus():
            client.add_handler(CallbackQueryHandler(
                pass_handler_and_clean(menu.on_callback, self),
                self.filter(menu.menu_id)))


def pass_handler_and_clean(func: Callable[[Client, Any], None],
                           handler: ParameterizedHandler) -> Callable[
                               [Client, Any], None]:
    def on_callback(client: Client, context):
        if isinstance(context, CallbackQuery):
            func(handler, client, context, context.parameters)
            handler.database.delete(context.parameters['callback_query_id'])
        else:
            func(handler, client, context)

    return on_callback



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

from .database.base_database import BaseDatabase
from .node import Node
from .parameterized_handler import ParameterizedHandler, pass_handler_and_clean
from .tree_handler import TreeHandler
from dataclasses import dataclass
from pyrogram import Client, MessageHandler


@dataclass(eq=False, init=False, repr=True)
class ParameterizedTreeHandler(TreeHandler, ParameterizedHandler):
    def __init__(self, main_node: Node,
                 database: BaseDatabase):
        TreeHandler.__init__(self, main_node)
        ParameterizedHandler.__init__(self, database)

    def setup(self, client: Client):
        ParameterizedHandler.setup(self, client)

        client.add_handler(
            MessageHandler(
                pass_handler_and_clean(
                    self.main_node.menu.on_message, self)))



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

from .base_menu import BaseMenu
from dataclasses import dataclass
from typing import List, Optional, Tuple


@dataclass(eq=False, init=False, repr=True)
class Node:
    menu: BaseMenu
    children: Optional[List['Node']] = None

    def __init__(self, menu: BaseMenu,
                 children: Optional[List['Node']] = None):
        if children:
            self.children = children
        else:
            self.children = []

        self.menu = menu

    def add_child(self, node: 'Node'):
        self.children.append(node)

    def get_children_menus(self) -> List[BaseMenu]:
        children = [child.menu for child in self.children]
        return children if children else None

    def get_family(self, menu_id: str,
                   parent: Optional['Node']) -> Tuple[Optional[BaseMenu],
                                                      Optional[
                                                          List[BaseMenu]]]:
        if self.menu.menu_id == menu_id:
            return (parent.menu if isinstance(parent, Node) else None,
                    self.get_children_menus())

        for child in self.children:
            child_menus = child.get_family(menu_id, self)

            if child_menus[0]:
                return child_menus

        return (None, None)

    def get_menus(self) -> List[BaseMenu]:
        menus = [self.menu]

        for child in self.children:
            if child.menu.menu_id != self.menu.menu_id:
                menus += child.get_menus()

        return menus



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

from .button import Button # noqa
from .database import DictDatabase, RedisDatabase # noqa
from .element import Element # noqa
from .keyboard import Keyboard # noqa
from .node import Node # noqa
from .page_menu import PageMenu # noqa
from .parameterized_tree_handler import ParameterizedTreeHandler # noqa
from .tree_menu import TreeMenu # noqa
from .tree_handler import TreeHandler, transform_dict # noqa

__author__ = "Hearot"
__author_email__ = "gabriel@hearot.it"
__license__ = "GNU General Public License v3"
__package__ = "pyrubrum"
__url__ = "https://github.com/hearot/pyrubrum"
__version__ = "0.1a0"



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

from .base_menu import BaseMenu
from .base_handler import BaseHandler, pass_handler
from .node import Node, Optional
from dataclasses import dataclass
from pyrogram import Client, MessageHandler
from typing import Any, Dict, Iterable, List, Tuple, Union


@dataclass(eq=False, init=False, repr=True)
class TreeHandler(BaseHandler):
    main_node: Node

    def __init__(self, main_node: Node):
        self.main_node = main_node

    def get_family(self, menu_id: str) -> Tuple[Optional[BaseMenu],
                                                Optional[List[BaseMenu]]]:
        return self.main_node.get_family(menu_id, None)

    def get_menus(self) -> List[BaseMenu]:
        return self.main_node.get_menus()

    def setup(self, client: Client):
        BaseHandler.setup(self, client)

        client.add_handler(
            MessageHandler(pass_handler(
                self.main_node.menu.on_message, self)))


def on_callback_node(menus: Union[Dict[BaseMenu, Any],
                                  Iterable[BaseMenu]],
                     parent: Node):
    for menu in menus:
        node = Node(menu)
        parent.add_child(node)

        if isinstance(menus, dict) and isinstance(menus[menu], Iterable):
            on_callback_node(menus[menu], node)


def transform_dict(menus: Union[Dict[BaseMenu, Any],
                                Iterable[BaseMenu]]) -> Node:
    main_node = Node(list(menus)[0])
    main_value = list(menus.values())[0] if isinstance(menus, dict) else None

    if main_value:
        on_callback_node(main_value, main_node)

    return main_node



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

from .base_handler import BaseHandler
from .button import Button
from dataclasses import dataclass
from pyrogram import (CallbackQuery, Client,
                      InlineKeyboardMarkup,
                      InputMedia, Message)
from typing import Any, Dict, Union


@dataclass(eq=False, init=False, repr=True)
class BaseMenu:
    name: str
    menu_id: str

    def __hash__(self):
        return hash(self.menu_id)

    def __init__(self, name: str, menu_id: str):
        self.name = name
        self.menu_id = menu_id

    def get_content(self) -> Union[InputMedia, str]:
        raise NotImplementedError

    def on_callback(self, handler: BaseHandler,
                    client: Client, callback: CallbackQuery,
                    parameters: Dict[str, Any]):
        raise NotImplementedError

    def button(self, handler: BaseHandler, client: Client,
               context: Union[CallbackQuery,
                              Message],
               parameters: Dict[str, Any]) -> Button:
        return Button(self.name, self.menu_id,
                      parameters)

    def keyboard(self, handler: BaseHandler, client: Client,
                 context: Union[CallbackQuery,
                                Message],
                 parameters: Dict[str, Any]) -> InlineKeyboardMarkup:
        raise NotImplementedError

    def on_message(self, handler: BaseHandler,
                   client: Client, message: Message,
                   parameters: Dict[str, Any]):
        raise NotImplementedError



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

from .button import Button
from .element import Element
from .keyboard import Keyboard
from .parameterized_tree_handler import ParameterizedTreeHandler
from .tree_menu import TreeMenu
from dataclasses import dataclass
from pyrogram import (CallbackQuery, Client,
                      InlineKeyboardMarkup,
                      InputMedia, Message)
from typing import Any, Callable, Dict, List, Optional, Union


@dataclass(eq=False, init=False, repr=True)
class PageMenu(TreeMenu):
    items: Union[List[Element],
                 Callable[[ParameterizedTreeHandler, Client,
                           Union[CallbackQuery, Message],
                           Dict[str, Any]],
                          List[Element]]]
    limit_page: Optional[int] = 4
    next_page_button_text: Optional[str] = "▶️"
    previous_page_button_text: Optional[str] = "◀️"

    def __init__(self, name: str,
                 menu_id: str,
                 content: Union[InputMedia, str],
                 items: Union[List[Element],
                              Callable[[ParameterizedTreeHandler, Client,
                                        Union[CallbackQuery, Message],
                                        Dict[str, Any]],
                                       List[Element]]],
                 back_button_text: Optional[str] = "🔙",
                 limit: Optional[int] = 2,
                 limit_page: Optional[int] = 4,
                 next_page_button_text: Optional[str] = "▶️",
                 previous_page_button_text: Optional[str] = "◀️"):
        TreeMenu.__init__(self, name, menu_id, content,
                          back_button_text=back_button_text,
                          limit=limit)

        self.items = items
        self.limit_page = limit_page
        self.next_page_button_text = next_page_button_text
        self.previous_page_button_text = previous_page_button_text

    def keyboard(self, tree: ParameterizedTreeHandler,
                 client: Client,
                 context: Union[CallbackQuery,
                                Message],
                 parameters: Dict[str, Any]) -> InlineKeyboardMarkup:
        parent, children = tree.get_family(self.menu_id)

        keyboard = []
        items = []
        page_id = 'page_' + self.menu_id

        if page_id not in parameters:
            parameters[page_id] = 0

        if 'same_menu' in parameters and parameters['same_menu']:
            parameters[page_id] = int(parameters['element_id'])

        if children:
            page_item_menu = children.pop(0)

            if callable(self.items):
                items = self.items(tree, client,
                                   context, parameters)
            elif isinstance(self.items, list):
                items = self.items
            else:
                raise TypeError("items must be either callable or a list")

            elements = items[
                parameters[page_id]*self.limit_page:][:self.limit_page]

            keyboard = [[Button(
                element.name, page_item_menu.menu_id,
                parameters,
                element.element_id) for element in
                        elements[i:i+self.limit]] for i in
                        range(0, len(elements), self.limit)]

        if children:
            keyboard += [[child.button(tree, client,
                                       context, parameters) for child in
                         children[i:i+self.limit]] for i in
                         range(0, len(children), self.limit)]

        teleport_row = []

        if parameters[page_id] > 0:
            previous_page_button = Button(
                self.previous_page_button_text,
                self.menu_id,
                parameters.copy(),
                parameters[page_id]-1,
                True)

            if self.menu_id + "_id" in parameters:
                previous_page_button.parameters[
                    self.menu_id + "_id"] = parameters[self.menu_id + "_id"]

            teleport_row.append(previous_page_button)

        if (parameters[page_id]+1)*self.limit_page < len(items):
            next_page_button = Button(
                self.next_page_button_text,
                self.menu_id,
                parameters.copy(),
                parameters[page_id]+1,
                True)

            if self.menu_id + "_id" in parameters:
                next_page_button.parameters[
                    self.menu_id + "_id"] = parameters[self.menu_id + "_id"]

            teleport_row.append(next_page_button)

        if teleport_row:
            keyboard += [teleport_row]

        if parent:
            parent_button = parent.button(tree, client, context, parameters)
            parent_button.name = self.back_button_text

            keyboard = keyboard + [[parent_button]]

        if isinstance(context, Message):
            return (Keyboard(keyboard, tree,
                    str(context.message_id) +
                    str(context.from_user.id))
                    if keyboard else None)
        elif isinstance(context, CallbackQuery):
            return (Keyboard(keyboard, tree,
                    context.id) if keyboard else None)



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import Any, Dict


@dataclass(init=False)
class Button:
    button_id: str
    element_id: str = ""
    name: str
    parameters: dict

    def __init__(self, name: str, button_id: str,
                 parameters: Dict[str, Any],
                 element_id: str = "",
                 same_menu: bool = False,
                 **kwargs):
        self.button_id = button_id
        self.element_id = element_id
        self.name = name
        self.parameters = parameters.copy()
        self.parameters.update(kwargs)

        self.parameters['button_id'] = button_id
        self.parameters['same_menu'] = same_menu



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

from .base_menu import BaseMenu
from .keyboard import Keyboard
from .tree_handler import TreeHandler
from dataclasses import dataclass
from pyrogram import (Client, CallbackQuery,
                      InlineKeyboardMarkup, InputMedia,
                      Message)
from typing import Any, Callable, Dict, Optional, Union


@dataclass(eq=False, init=False, repr=True)
class TreeMenu(BaseMenu):
    content: Union[Union[InputMedia, str],
                   Callable[[TreeHandler, Client,
                             Union[CallbackQuery, Message],
                             Dict[str, Any]],
                            Union[InputMedia, str]]]
    back_button_text: Optional[str] = "🔙"
    limit: Optional[int] = 2

    def __init__(self, name: str, menu_id: str,
                 content: Union[Union[InputMedia, str],
                                Callable[[TreeHandler, Client,
                                          Union[CallbackQuery, Message],
                                          Dict[str, Any]],
                                         Union[InputMedia, str]]],
                 back_button_text: Optional[str] = "🔙",
                 limit: Optional[int] = 2):
        BaseMenu.__init__(self, name, menu_id)
        self.back_button_text = back_button_text
        self.content = content
        self.limit = limit

    def get_content(self, tree: TreeHandler, client: Client,
                    context: Union[CallbackQuery,
                                   Message],
                    parameters: Dict[str, Any]) -> Union[InputMedia, str]:
        if callable(self.content):
            return self.content(tree, client, context, parameters)

        return self.content

    def preliminary(self, tree: TreeHandler, client: Client,
                    context: Union[CallbackQuery, Message],
                    parameters: Dict[str, Any]):
        pass

    def on_callback(self, tree: TreeHandler, client: Client,
                    callback: CallbackQuery,
                    parameters: Dict[str, Any]):
        self.preliminary(tree, client, callback, parameters)
        content = self.get_content(tree, client, callback, parameters)

        if isinstance(content, InputMedia):
            callback.edit_message_media(
                content, reply_markup=self.keyboard(
                    tree, client, callback, parameters))
        elif isinstance(content, str):
            callback.edit_message_text(
                content,
                reply_markup=self.keyboard(
                    tree, client, callback, parameters))
        else:
            raise TypeError("content must be of type InputMedia or str")

    def keyboard(self, tree: TreeHandler, client: Client,
                 context: Union[CallbackQuery,
                                Message],
                 parameters: Dict[str, Any]) -> InlineKeyboardMarkup:
        parent, children = tree.get_family(self.menu_id)

        keyboard = []

        if children:
            keyboard = [[child.button(tree, client,
                                      context, parameters) for child in
                        children[i:i+self.limit]] for i in
                        range(0, len(children), self.limit)]

        if parent:
            parent_button = parent.button(tree, client, context, parameters)
            parent_button.name = self.back_button_text

            keyboard = keyboard + [[parent_button]]

        if isinstance(context, Message):
            return (Keyboard(keyboard, tree,
                    str(context.message_id) +
                    str(context.from_user.id))
                    if keyboard else None)
        elif isinstance(context, CallbackQuery):
            return (Keyboard(keyboard, tree,
                    context.id) if keyboard else None)

    def on_message(self, tree: TreeHandler, client: Client,
                   message: Message):
        parameters = {}

        self.preliminary(tree, client, message, parameters)
        content = self.get_content(tree, client, message, parameters)

        if isinstance(content, InputMedia):
            raise NotImplementedError  # TODO: handle media
        elif isinstance(content, str):
            message.reply_text(content,
                               reply_markup=self.keyboard(
                                   tree, client,
                                   message, parameters
                               ))
        else:
            raise TypeError("content must be of type InputMedia or str")
